function love.load()
	state = "menu"

	-- load graphics
	tileset = love.graphics.newImage('crystals.png')

	-- set up tilesheet and quads
	tileW, tileH = 64,64
	local tilesetW, tilesetH = tileset:getWidth(), tileset:getHeight()
	local numRows,numCols = tilesetH / tileH, tilesetW / tileW

	local quadInfo = {}
	for tRowIndex = 1, numRows do
		quadInfo[tRowIndex] = {}
		for tColIndex = 1, numCols do
			quadInfo[tRowIndex][tColIndex] = { (tColIndex - 1) * tileW, (tRowIndex - 1) * tileH }
		end
	end

	Quads = {}
	for i, tCol in ipairs(quadInfo) do
		for j, info in ipairs(tCol) do
			Quads[(i - 1) * numCols + j] = love.graphics.newQuad(info[1], info[2], tileW, tileH, tilesetW, tilesetH)
		end
	end

	-- set screen size
	gridRows = 13
	gridCols = 8
	love.graphics.setMode(tileW * gridCols, tileH * gridRows)
	screenWidth, screenHeight = love.graphics.getWidth(), love.graphics.getHeight()

	-- set up save
	love.filesystem.setIdentity("match3builder")

	-- set up highscore
	highscore = require 'sick'
	highscore.set("scores.txt", 5, "abc", 0)

	-- set audio
	local music = love.audio.newSource("music.mp3","stream")
	music:setLooping(true)
	love.audio.play(music)

	sound = love.audio.newSource("match.mp3","static")

	-- set up fonts
	normalFont = love.graphics.newFont("imagine_font.ttf", 14)
	largeFont = love.graphics.newFont("imagine_font.ttf", 30)
	love.graphics.setFont(normalFont)
end

function setup()
	state = "game"
	newHighScore = false

	-- set up grid
	grid = {}
	for rowIndex = gridRows, 1, -1 do
		grid[rowIndex] = {}
	end

	repeat
		for row = gridRows, 8, -1 do
			for col = 1, gridCols do
				newP = newPiece(row, col)
				newP.y = (row - 1) * tileH
				newP.isDropping = false
			end
		end
	until (next(lookForMatches()) == nil)

	-- init time counter
	timePassed = 0
	difficultyTimer = 0
	spawnNumTimer = 0
	spawnNum = 1
	delay = 1

	-- init score
	score = 0

	
end

function love.update(dt)

	if state == "game" then
		-- check if any piece has landed above the line
		for i = 1, gridCols do
			if grid[1][i] ~= nil then
				if grid[1][i].isDropping == false then
					gameOver()
					break
				end
			end
		end

		-- update timers
		timePassed = timePassed + dt
		difficultyTimer = difficultyTimer + dt
		spawnNumTimer = spawnNumTimer + dt


		-- increase difficulty
		if difficultyTimer > 10 then
			delay = delay - 0.1
			if delay < 0.4 then
				delay = 0.4
			end
			difficultyTimer = 0
		end
	
		if spawnNumTimer > 30 then
			--spawnNum = spawnNum + 1
			spawnNumTimer = 0
		end
	
		if swapTimer then
			swapTimer = swapTimer + dt
		end

		movePieces()
		if timePassed > delay then
			for numBlocks = 1, spawnNum do
				local x = math.random(1,gridCols)
				local y
				for i = gridRows,1,-1 do
					if grid[i][x] == nil then
						y = i
						break
					end
				end
				if y == nil then break end
				newPiece(y, x)
				timePassed = 0
			end
		end
	end
end

function love.draw()
	if state == "menu" then
		love.graphics.setFont(largeFont)
		love.graphics.printf("OVERFLOW", 0, screenHeight / 2 - 30, screenWidth, "center")
		love.graphics.setFont(normalFont)
		love.graphics.printf("Match same colors to clear\n\nStay below the line\n\nClick to Begin", 0, screenHeight / 2 , screenWidth, "center")
	elseif state == "game" then
		love.graphics.line(0,tileH, tileW * gridCols, tileH)
		for row = gridRows,1,-1 do
			for col = 1, gridCols do
				piece = grid[row][col]
				if piece then
      					love.graphics.drawq(tileset, Quads[piece.type], piece.x, piece.y)
				end
			end
		end
		
		-- draw rectangle around selected piece
		if firstPiece then
			love.graphics.setColor(255,255,255,255)
			love.graphics.rectangle("line", (firstPiece.col-1) * tileW, (firstPiece.row-1) * tileH, tileW, tileH)
			love.graphics.setColor(255,255,255,255)
		end

		-- draw score
		hscore = 0
		for i, score, name in highscore() do
			hscore = score
			break
		end
		love.graphics.printf("HIGHSCORE: " ..hscore, 0,5, screenWidth, "left")
		love.graphics.printf("SCORE: " ..score,0,5, screenWidth, "right")

	elseif state == "game over" then
		love.graphics.setFont(largeFont)
		love.graphics.printf(score, 0, screenHeight / 2 - 35, screenWidth, "center")
		love.graphics.setFont(normalFont)
		if newHighScore then
			love.graphics.printf("NEW HIGHSCORE\n\nClick to Restart", 0, screenHeight / 2, screenWidth, "center")
		else
			love.graphics.printf("GAME OVER\n\nClick to Restart", 0, screenHeight / 2, screenWidth, "center")
		end
	end
end

function lookForMatches()
	local matchList = {}

	-- search for horizontal matches
	for row = 1, gridRows do
		for col = 1, gridCols do
			local match = getMatchHoriz(row, col)
			if (#match > 2) then
				table.insert(matchList, match)
				--col = col + #match - 1 -- skip blocks in the match
			end
		end
	end
			
	-- search for vertical matches
	for col = 1, gridCols do
		for row = 1, gridRows do
			local match = getMatchVert(row, col)
			if (#match > 2) then
				table.insert(matchList, match)
				--row = row + #match - 1 -- skip blocks in the match
			end
		end
	end

	return matchList
end

function getMatchHoriz(row, col)
	local match = {grid[row][col]}
	for i = 1, gridCols - col do
		if grid[row][col] and grid[row][col].isDropping == false  then
			if grid[row][col+i] == nil then
				return match
			end
			if grid[row][col].type == grid[row][col + i].type and grid[row][col+i].isDropping == false then
				table.insert(match, grid[row][col + i])
			else
				return match
			end
		end
	end
	return match
end
		
function getMatchVert(row,col)
	local match = {grid[row][col]}
	for i = 1, gridRows - row do
		if grid[row][col] and grid[row][col].isDropping == false then
			if grid[row+i][col] == nil then
				return match
			end
			if grid[row][col].type == grid[row + i][col].type and grid[row+i][col].isDropping == false then
				table.insert(match, grid[row + i][col])
			else
				return match
			end
		end
	end
	return match
end

function lookForPossibles()
	for row = 1, 8 do
		-- horizontal possible two plus one
		if matchPattern(col, row, {{1,0}}, {{-2,0},{-1,-1},{-1,1},{2,-1},{2,1},{3,0}}) then
			return true 
		end

		-- horizontal possible middle
		if matchPattern(col, row, {{2,0}},{{1,-1},{1,1}}) then
			return true
		end

		-- vertical possible two plus one
		if matchPattern(col, row, {{0,1}}, {{0,-2},{-1,-1},{1,-1},{-1,2},{1,2},{0,3}}) then
			return true
		end

		-- vertical possible middle
		if matchPattern(col, row, {{0,2}}, {{-1,1},{1,1}}) then
			return true
		end
	end
end

function matchPattern(col, row, mustHave, needOne)
	thisType = grid[col][row].type
	for i = 1, #mustHave do
		if not matchType(col + mustHave[i][1], row + mustHave[i][2], thisType) then
			return false
		end
	end

	for i = 1, #needOne do
		if matchType(col + needOne[i][1], row + needOne[i][2], thisType) then
			return true
		end
	end

	return false
end

function matchType(col, row, type)
	if (col < 0) or (col > 7) or (row < 0) or (row > 7) then
		return false
	end

	return (grid[col][row].type == type)
end

function love.mousereleased(x,y,button)
	if state == "menu" then
		setup()
	elseif state == "game" then
		for row = 1, gridRows do
			for col = 1, gridCols do
				if (x > (col - 1) * tileW) and (x < (col * tileW)) and (y > (row - 1) * tileH) and (y < (row * tileH)) then
					local clickedPiece
					if grid[row][col] ~= nil then
						clickedPiece = grid[row][col]
					end
					if clickedPiece == nil then
						firstPiece = nil

					elseif not firstPiece and clickedPiece.isDropping == false then
						firstPiece = clickedPiece
					elseif firstPiece == clickedPiece or clickedPiece.isDropping then
						firstPiece = nil
					else
						-- same row, one column over
						if firstPiece.row == clickedPiece.row and math.abs(firstPiece.col - clickedPiece.col) == 1 then
							makeSwap(firstPiece, clickedPiece)
							firstPiece = nil
						-- same col, one row over
						elseif firstPiece.col == clickedPiece.col and math.abs(firstPiece.row - clickedPiece.row) == 1 then
							makeSwap(firstPiece, clickedPiece)
							firstPiece = nil
						else
							firstPiece = clickedPiece
						end
					end
					break
				end
			end
		end
	elseif state == "game over" then
		setup()
	end
end

function makeSwap(piece1, piece2)
	swapPieces(piece1, piece2)

	-- if no match
	if next(lookForMatches()) == nil then
		swapPieces(piece1,piece2)
	else
		isSwapping = true
		swapTimer = 0
	end
end

function swapPieces(piece1,piece2)
	tmpCol = piece1.col;
	tmpRow = piece1.row;
	piece1.col = piece2.col;
	piece1.row = piece2.row;
	piece2.col = tmpCol;
	piece2.row = tmpRow;

	grid[piece1.row][piece1.col]= piece1;
	grid[piece2.row][piece2.col] = piece2;

end

function movePieces()
	madeMove = false
	for row = 1, gridRows do
		for col = 1, gridCols do
			if grid[row][col] then
				--needs to move down
				if grid[row][col].y < (grid[row][col].row - 1)* tileH then
					grid[row][col].y = grid[row][col].y + tileH / 16
					madeMove = true

					--needs to move up
				elseif grid[row][col].y > (grid[row][col].row - 1) * tileH then
					grid[row][col].y = grid[row][col].y - tileH / 16
					madeMove = true;

					--needs to move right
				elseif grid[row][col].x < (grid[row][col].col - 1) * tileW then
					grid[row][col].x = grid[row][col].x + tileW / 16
					madeMove = true;

					--needs to move left
				elseif grid[row][col].x > (grid[row][col].col - 1) * tileW then
					grid[row][col].x = grid[row][col].x - tileW / 16
					madeMove = true;
				else
					grid[row][col].isDropping = false
				end
			end
		end
	end

	if swapTimer == nil then
		findAndRemoveMatches()
	elseif swapTimer > 0.4 then
		findAndRemoveMatches()
		isSwapping = false
		swapTimer = nil
	end
end

function findAndRemoveMatches()
	local matches = lookForMatches()
	for i=1, #matches do
		for j=1, #matches[i] do
			grid[matches[i][j].row][matches[i][j].col] = nil
			affectAbove(matches[i][j])
			score = score + #matches * 10
			love.audio.play(sound)
		end
	end
end

function affectAbove(piece)
	for row = piece.row - 1, 1, -1 do
		if grid[row][piece.col] then
			grid[row][piece.col].row = grid[row][piece.col].row + 1
			grid[row][piece.col].isDropping = true
			grid[row + 1][piece.col] = grid[row][piece.col]
			grid[row][piece.col] = nil
		end
	end
end

function newPiece(rowIndex, colIndex)
	piece = {}
	piece.type = math.random(16,19)
	piece.col = colIndex
	piece.row = rowIndex
	piece.x, piece.y = (colIndex-1)*tileW, (rowIndex-1)*tileH - tileH * gridRows
	grid[rowIndex][colIndex] = piece
	piece.isDropping = true
	return piece
end

function gameOver()
	highscore.save()
	state = "game over"
	highscore.add("player",score)
	if hscore == score then
		newHighScore = true
	end
end
